import 'dart:async';
import 'dart:math';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:fotsporquiz/screens/HomeScreen_screen.dart';
import 'package:fotsporquiz/screens/PostIntroduction_screen.dart';
import 'package:fotsporquiz/screens/Question_screen.dart';
import 'package:fotsporquiz/services/quiz.service.dart';
import 'package:google_maps_flutter/google_maps_flutter.dart';
import 'package:back_button_interceptor/back_button_interceptor.dart';
import 'package:geolocator/geolocator.dart';
import 'package:label_marker/label_marker.dart';
import 'package:provider/provider.dart';
import '../entities/post.dart';
import '../entities/quiz.dart';
import '../providers/quiz_provider.dart';

class MapScreen extends StatefulWidget {
  const MapScreen({Key? key}) : super(key: key);

  @override
  _MapScreenState createState() => _MapScreenState();
}

class _MapScreenState extends State<MapScreen> {
  LatLng currentLatLng = const LatLng(60.397076, 5.324383); // Player
  LatLng initialPosition = const LatLng(0, 0); // Camera
  final Set<Marker> _markerList = {};
  late List<Post> _posts = [];
  final QuizService _quizService = QuizService();
  GoogleMapController? _googleMapController;
  Timer? timer;

  bool isMapVisible = false;

  @override
  void initState() {
    super.initState();

    BackButtonInterceptor.add(myInterceptor);
    timer = Timer.periodic(const Duration(seconds: 1), ((timer) {
      _currentPosition().then((currLocation) {
        currentLatLng = LatLng(currLocation.latitude, currLocation.longitude);
      });
      checkPosition(_posts);
    }));
  }

  @override
  void dispose() {
    super.dispose();
    _googleMapController?.dispose();
    _markerList.clear();
    BackButtonInterceptor.remove(myInterceptor);
  }

  LatLngBounds _bounds(Set<Marker> markers) {
    return _createBounds(markers.map((m) => m.position).toList());
  }

  LatLngBounds _createBounds(List<LatLng> positions) {
    final southwestLat = positions.map((p) => p.latitude).reduce(
        (value, element) => value < element ? value : element); // smallest
    final southwestLon = positions
        .map((p) => p.longitude)
        .reduce((value, element) => value < element ? value : element);
    final northeastLat = positions.map((p) => p.latitude).reduce(
        (value, element) => value > element ? value : element); // biggest
    final northeastLon = positions
        .map((p) => p.longitude)
        .reduce((value, element) => value > element ? value : element);
    return LatLngBounds(
        southwest: LatLng(southwestLat, southwestLon),
        northeast: LatLng(northeastLat, northeastLon));
  }

  // Checks if player is within the radius of a post, and if so opens it.
  checkPosition(List<Post> posts) {
    LatLng currLocation =
        LatLng(currentLatLng.latitude, currentLatLng.longitude);
    for (Post p in posts) {
      if (!p.isAnswered) {
        if (calculateDistance(
                currLocation, LatLng(p.position[0], p.position[1])) <
            15 /* <- radius */) {
          // <- tallet som er radius
          if (p.description == "") {
            Navigator.pushReplacement(
              context,
              MaterialPageRoute(
                builder: (context) => QuestionScreen(p),
              ),
            );
          } else {
            Navigator.pushReplacement(
              context,
              MaterialPageRoute(
                builder: (context) => PostIntroductionScreen(p),
              ),
            );
          }
        }
      }
    }
  }

  // Calculates the distance between two points
  double calculateDistance(LatLng pos1, LatLng pos2) {
    return Geolocator.distanceBetween(
        pos1.latitude, pos1.longitude, pos2.latitude, pos2.longitude);
  }

  // Refuses back button to activate the back button in the MapScreen
  bool myInterceptor(bool stopDefaultButtonEvent, RouteInfo info) {
    return true;
  }

  // Generates a list of posts, which gets displayed on the map
  void generateMarkers(GameProvider p) async {
    for (var i = 0; i < p.quiz.posts.length; i++) {
      var post = p.quiz.posts[i];
      if (!post.isAnswered) {
        _markerList.addLabelMarker(
          LabelMarker(
            textStyle:
                const TextStyle(fontSize: 50.0, fontWeight: FontWeight.bold),
            label: post.id.toString(),
            markerId: (MarkerId(post.id.toString())),
            infoWindow: InfoWindow(title: 'Post ' + post.id.toString()),
            backgroundColor: Colors.black,
            position: LatLng(post.position.first, post.position.last),
          ),
        );
      } else {
        _markerList.addLabelMarker(
          LabelMarker(
            textStyle: const TextStyle(
              fontSize: 50.0,
              fontWeight: FontWeight.bold,
            ),
            label: post.id.toString(),
            markerId: (MarkerId(post.id.toString())),
            infoWindow: InfoWindow(title: 'Post ' + post.id.toString()),
            backgroundColor: Colors.black38,
            position: LatLng(post.position.first, post.position.last),
          ),
        );
      }
    }
  }

  // Check if all posts are visited
  bool outOfPosts(Quiz quiz) {
    bool isDone = true;
    for (Post p in quiz.posts) {
      if (!p.isAnswered) {
        isDone = false;
      }
    }
    return isDone;
  }

  // Returns to the home screen
  void returnToHomeScreen(GameProvider p) {
    p.reset();
    timer?.cancel();
    Navigator.pushReplacement(
      context,
      MaterialPageRoute(
        builder: (context) => HomeScreen(),
      ),
    );

    for (Post p in p.quiz.posts) {
      p.isAnswered = false;
    }
  }

  // Converts a list of posts to a LatLng object
  LatLng converter(List<double> pos) {
    return LatLng(pos[0], pos[1]);
  }

  bool shouldPop = true;

  @override
  Widget build(BuildContext context) {
    SystemChrome.setPreferredOrientations([
      DeviceOrientation.portraitUp,
      DeviceOrientation.portraitDown,
    ]);
    var gameProvider = Provider.of<GameProvider>(context);

    generateMarkers(gameProvider);
    _posts = gameProvider.quiz.posts;

    return WillPopScope(
      onWillPop: () async {
        return shouldPop;
      },
      child: Scaffold(
        appBar: AppBar(
          title: Text(
            gameProvider.player.nickname.toUpperCase() +
                " : " +
                gameProvider.player.score.toString() +
                " Poeng",
            style: const TextStyle(
              fontSize: 18,
              color: Colors.white,
            ),
          ),
          centerTitle: true,
          elevation: 8,
          leading: IconButton(
            icon: const Icon(Icons.refresh),
            onPressed: () {
              Navigator.pushReplacement(
                context,
                MaterialPageRoute(
                  builder: (context) => const MapScreen(),
                ),
              );
            },
          ),
          actions: [
            IconButton(
              onPressed: (() {
                showDialog(
                  context: context,
                  builder: (context) => AlertDialog(
                    title: const Text("Introduksjon til rute:"),
                    content: Text(gameProvider.quiz.description),
                    actions: [
                      TextButton(
                        onPressed: () {
                          Navigator.pop(context);
                        },
                        child: const Text("Lukk"),
                      ),
                    ],
                  ),
                );
              }),
              icon: const Icon(Icons.info),
            ),
            IconButton(
              icon: const Icon(Icons.home),
              onPressed: () {
                showDialog(
                  context: context,
                  builder: (context) => AlertDialog(
                    title: const Text("Er du sikker?"),
                    content: const Text(
                        "Du vil miste resultatene dine hvis du går ut av kartet."),
                    actions: [
                      TextButton(
                        onPressed: () {
                          returnToHomeScreen(gameProvider);
                        },
                        child: const Text("Ja"),
                      ),
                      TextButton(
                        onPressed: () {
                          Navigator.pop(context);
                        },
                        child: const Text("Nei"),
                      )
                    ],
                  ),
                );
              },
            ),
          ],
          //backgroundColor: const Color(0xFFBBDC5B),
          flexibleSpace: Container(
            decoration: const BoxDecoration(
              gradient: LinearGradient(
                colors: [Color.fromRGBO(186, 220, 91, 1), Colors.blueGrey],
              ),
            ),
          ),
          systemOverlayStyle: SystemUiOverlayStyle.light,
          automaticallyImplyLeading: false,
        ),
        body: !outOfPosts(gameProvider.quiz)
            ? AnimatedOpacity(
                curve: Curves.fastOutSlowIn,
                opacity: isMapVisible ? 1.0 : 0,
                duration: const Duration(milliseconds: 600),
                child: GoogleMap(
                  initialCameraPosition: CameraPosition(
                      target: converter(gameProvider.quiz.posts.first.position),
                      zoom: 13),
                  onMapCreated: (controller) => {
                    _googleMapController = controller,
                    setState(() {
                      controller.animateCamera(CameraUpdate.newLatLngBounds(
                          _bounds(_markerList), 50));
                      Future.delayed(
                          const Duration(milliseconds: 550),
                          () => setState(() {
                                isMapVisible = true;
                              }));
                    }),
                  },
                  zoomControlsEnabled: false,
                  myLocationEnabled: true,
                  rotateGesturesEnabled: true,
                  mapToolbarEnabled: false,
                  markers: Set<Marker>.of(_markerList),
                ),
              )
            : Center(
                child: Column(
                  mainAxisAlignment: MainAxisAlignment.center,
                  crossAxisAlignment: CrossAxisAlignment.center,
                  children: [
                    const Text(
                      "Tom for poster.",
                      style: TextStyle(fontSize: 35),
                    ),
                    const SizedBox(
                      height: 30,
                    ),
                    Text(
                      "Ditt resultat:  " +
                          gameProvider.player.score.toString() +
                          "/" +
                          _quizService
                              .countQuestions(gameProvider.quiz)
                              .toString(),
                      style: const TextStyle(
                          fontSize: 30, fontWeight: FontWeight.bold),
                    ),
                    SizedBox(
                      width: double.infinity,
                      height: 100,
                      child: Padding(
                        padding: const EdgeInsets.symmetric(
                            horizontal: 15, vertical: 15),
                        child: ElevatedButton(
                          onPressed: () {
                            returnToHomeScreen(gameProvider);
                          },
                          child: const Text("Returner til hovedmeny."),
                          style: ButtonStyle(
                            shape: MaterialStateProperty.all<
                                RoundedRectangleBorder>(RoundedRectangleBorder(
                              borderRadius: BorderRadius.circular(12.0),
                            )),
                            backgroundColor:
                                MaterialStateProperty.resolveWith<Color>(
                              (Set<MaterialState> states) {
                                if (states.contains(MaterialState.pressed)) {
                                  return const Color(0xFFBBDC5B);
                                }
                                return Colors.black87;
                              },
                            ),
                          ),
                        ),
                      ),
                    ),
                  ],
                ),
              ),
      ),
    );
  }

  Future<Position> _currentPosition() async {
    bool serviceEnabled;
    LocationPermission permission;
    serviceEnabled = await Geolocator.isLocationServiceEnabled();
    if (!serviceEnabled) {
      return Future.error('Location services are disabled');
    }
    permission = await Geolocator.checkPermission();
    if (permission == LocationPermission.denied) {
      permission = await Geolocator.requestPermission();
      if (permission == LocationPermission.denied) {
        return Future.error("Location permission denied");
      }
    }
    if (permission == LocationPermission.deniedForever) {
      return Future.error('Location permissions are permanently denied');
    }
    Position position = await Geolocator.getCurrentPosition();
    return position;
  }
}
